package org.apache.maven.doxia;

/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

import org.apache.maven.doxia.parser.manager.ParserNotFoundException;
import org.codehaus.plexus.PlexusTestCase;
import org.junit.Test;

public class DefaultDoxiaTest extends PlexusTestCase
{

    @Test
    public void testCreatesDefaultDoxia()
    {
        final DefaultDoxia defaultDoxia = new DefaultDoxia();

        assertNotNull( defaultDoxia );
    }

    @Test
    public void testFailsWhenParserIdDoesNotExist() throws Exception
    {
        final String parserId = "a-parser";
        final Doxia doxia = lookup( Doxia.class );

        try
        {
            doxia.getParser( parserId );
            fail( "Call should fail with ParserNotFoundException" );
        }
        catch ( ParserNotFoundException e )
        {
            assertEquals( "Cannot find parser with id = a-parser", e.getMessage() );
        }
    }

}
